/*
	Free view camera script
	--------------------------------------------------
	A script to control the camera in free view mode.
	Could be useful for a spectate mode for example.
	
	Author: Juppi
*/

pCamera <- null; // Camera class object
pLabel <- null; // Camera position label

class CCamera
{
	/// -- These are the interface funcs for this class
	// CCamera::Enable() - Enables the free camera
	// CCamera::Disable() - Disables the free camera
	// CCamera::IsEnabled() - Returns whether the camera has been enabled/disabled
	// CCamera::Process() - Processes the camera each frame and updates the GTA camera position
	
	/// -- These are general camera movement functions. They don't do anything else than calculate and update the internal member vars
	// CCamera::Rotate( fHoriz, fVert ) - Rotates the camera horizontally/vertically
	// CCamera::Move( fDist ) - Moves the camera forward/backward
	// CCamera::MoveSideways( fDist ) - Moves the camera sideways
	
	function IsEnabled() { return bEnabled; }
	function _typeof() { return "CCamera"; }
	
	// Camera position
	vPos = Vector( 0.0, 0.0, 0.0 );
	vTarget = Vector( 0.0, 0.0, 0.0 );
	fYaw = 0.0;
	fPitch = 0.0;
	fSpeed = 1.5;
	
	// Camera status
	bEnabled = false;
	bMovingForward = false;
	bMovingLeft = false;
	bMovingBackward = false;
	bMovingRight = false;
	bRotatingUp = false;
	bRotatingLeft = false;
	bRotatingDown = false;
	bRotatingRight = false;
}

function CCamera::Enable()
{
	// Camera movement keys
	::BindKey( 'W', ::BINDTYPE_DOWN, ::HandleKeyDown, 'W' );
	::BindKey( 'A', ::BINDTYPE_DOWN, ::HandleKeyDown, 'A' );
	::BindKey( 'S', ::BINDTYPE_DOWN, ::HandleKeyDown, 'S' );
	::BindKey( 'D', ::BINDTYPE_DOWN, ::HandleKeyDown, 'D' );
	::BindKey( 'W', ::BINDTYPE_UP, ::HandleKeyUp, 'W' );
	::BindKey( 'A', ::BINDTYPE_UP, ::HandleKeyUp, 'A' );
	::BindKey( 'S', ::BINDTYPE_UP, ::HandleKeyUp, 'S' );
	::BindKey( 'D', ::BINDTYPE_UP, ::HandleKeyUp, 'D' );
	
	// Camera rotation keys
	::BindKey( ::KEY_UP, ::BINDTYPE_DOWN, ::HandleKeyDown, ::KEY_UP );
	::BindKey( ::KEY_LEFT, ::BINDTYPE_DOWN, ::HandleKeyDown, ::KEY_LEFT );
	::BindKey( ::KEY_DOWN, ::BINDTYPE_DOWN, ::HandleKeyDown, ::KEY_DOWN );
	::BindKey( ::KEY_RIGHT, ::BINDTYPE_DOWN, ::HandleKeyDown, ::KEY_RIGHT );
	::BindKey( ::KEY_UP, ::BINDTYPE_UP, ::HandleKeyUp, ::KEY_UP );
	::BindKey( ::KEY_LEFT, ::BINDTYPE_UP, ::HandleKeyUp, ::KEY_LEFT );
	::BindKey( ::KEY_DOWN, ::BINDTYPE_UP, ::HandleKeyUp, ::KEY_DOWN );
	::BindKey( ::KEY_RIGHT, ::BINDTYPE_UP, ::HandleKeyUp, ::KEY_RIGHT );
	
	// Set the initial place and rotation of the camera
	local pLocal = ::FindLocalPlayer();
	vPos = pLocal.Pos;
	fPitch = 0.0;
	fYaw = 0.0;
	
	Rotate( -1 * ::FindLocalPlayer().Angle * PI/180.0, 0.0 );
	
	::SetCameraMatrix( vPos, vTarget );
	pLocal.Frozen = true;
	
	::pLabel.Visible = true;
	::pLabel.Text = format( "Camera pos: %.1f, %.1f, %.1f", vPos.x, vPos.y, vPos.z );
	
	bEnabled = true;
}

function CCamera::Disable()
{
	// Camera movement keys
	::UnbindKey( 'W', ::BINDTYPE_DOWN, ::HandleKeyDown );
	::UnbindKey( 'A', ::BINDTYPE_DOWN, ::HandleKeyDown );
	::UnbindKey( 'S', ::BINDTYPE_DOWN, ::HandleKeyDown );
	::UnbindKey( 'D', ::BINDTYPE_DOWN, ::HandleKeyDown );
	::UnbindKey( 'W', ::BINDTYPE_UP, ::HandleKeyUp );
	::UnbindKey( 'A', ::BINDTYPE_UP, ::HandleKeyUp );
	::UnbindKey( 'S', ::BINDTYPE_UP, ::HandleKeyUp );
	::UnbindKey( 'D', ::BINDTYPE_UP, ::HandleKeyUp );
	
	// Camera rotation keys
	::UnbindKey( ::KEY_UP, ::BINDTYPE_DOWN, ::HandleKeyDown );
	::UnbindKey( ::KEY_LEFT, ::BINDTYPE_DOWN, ::HandleKeyDown );
	::UnbindKey( ::KEY_DOWN, ::BINDTYPE_DOWN, ::HandleKeyDown );
	::UnbindKey( ::KEY_RIGHT, ::BINDTYPE_DOWN, ::HandleKeyDown );
	::UnbindKey( ::KEY_UP, ::BINDTYPE_UP, ::HandleKeyUp );
	::UnbindKey( ::KEY_LEFT, ::BINDTYPE_UP, ::HandleKeyUp );
	::UnbindKey( ::KEY_DOWN, ::BINDTYPE_UP, ::HandleKeyUp );
	::UnbindKey( ::KEY_RIGHT, ::BINDTYPE_UP, ::HandleKeyUp );
	
	::pLabel.Visible = false;
	::pLabel.Text = "Camera pos: 0.0, 0.0, 0.0";
	
	local pLocal = ::FindLocalPlayer();
	pLocal.Pos = vPos;
	pLocal.Frozen = false;
	
	::RestoreCamera();
	
	bEnabled = false;
}

function CCamera::Rotate( fHoriz, fVert )
{
	fYaw += fHoriz;
	fPitch += fVert;
	
	vTarget.x = vPos.x + cos( fPitch ) * sin( fYaw );
	vTarget.y = vPos.y + cos( fPitch ) * cos( fYaw );
	vTarget.z = vPos.z + sin( fPitch );
}

function CCamera::Move( fDist )
{
	vPos.x += fDist * cos( fPitch ) * sin( fYaw );
	vPos.y += fDist * cos( fPitch ) * cos( fYaw );
	vPos.z += fDist * sin( fPitch );
	
	Rotate( 0.0, 0.0 );
}

function CCamera::MoveSideways( fDist )
{
	vPos.x += fDist * cos( fPitch ) * sin( fYaw + PI/2 );
	vPos.y += fDist * cos( fPitch ) * cos( fYaw + PI/2 );
	
	Rotate( 0.0, 0.0 );
}

function CCamera::Process()
{
	local bMoving = false;
	local fCamSpeed = ::GetKeyState( ::KEY_SHIFT ) ? 2.5 * fSpeed : fSpeed;
	
	if ( bMovingForward )
	{
		Move( fCamSpeed );
		bMoving = true;
	}
	if ( bMovingBackward )
	{
		Move( -fCamSpeed );
		bMoving = true;
	}
	if ( bMovingLeft )
	{
		MoveSideways( -fCamSpeed );
		bMoving = true;
	}
	if ( bMovingRight )
	{
		MoveSideways( fCamSpeed );
		bMoving = true;
	}
	if ( bRotatingUp )
	{
		Rotate( 0.0, 0.1 );
		bMoving = true;
	}
	if ( bRotatingDown )
	{
		Rotate( 0.0, -0.1 );
		bMoving = true;
	}
	if ( bRotatingLeft )
	{
		Rotate( -0.1, 0.0 );
		bMoving = true;
	}
	if ( bRotatingRight )
	{
		Rotate( 0.1, 0.0 );
		bMoving = true;
	}
	
	if ( bMoving )
	{
		::SetCameraMatrix( vPos, vTarget );
		::pLabel.Text = format( "Camera pos: %.1f, %.1f, %.1f", vPos.x, vPos.y, vPos.z );
	}
}

function HandleKeyDown( iKey )
{
	switch ( iKey )
	{
		case 'W':
			pCamera.bMovingForward = true;
			break;
		case 'A':
			pCamera.bMovingLeft = true;
			break;
		case 'S':
			pCamera.bMovingBackward = true;
			break;
		case 'D':
			pCamera.bMovingRight = true;
			break;
		case ::KEY_UP:
			pCamera.bRotatingUp = true;
			break;
		case ::KEY_LEFT:
			pCamera.bRotatingLeft = true;
			break;
		case ::KEY_DOWN:
			pCamera.bRotatingDown = true;
			break;
		case ::KEY_RIGHT:
			pCamera.bRotatingRight = true;
			break;
	}
}

function HandleKeyUp( iKey )
{
	switch ( iKey )
	{
		case 'W':
			pCamera.bMovingForward = false;
			break;
		case 'A':
			pCamera.bMovingLeft = false;
			break;
		case 'S':
			pCamera.bMovingBackward = false;
			break;
		case 'D':
			pCamera.bMovingRight = false;
			break;
		case ::KEY_UP:
			pCamera.bRotatingUp = false;
			break;
		case ::KEY_LEFT:
			pCamera.bRotatingLeft = false;
			break;
		case ::KEY_DOWN:
			pCamera.bRotatingDown = false;
			break;
		case ::KEY_RIGHT:
			pCamera.bRotatingRight = false;
			break;
	}
}


//--------------------------------------------------
// LU Events
//--------------------------------------------------

function onScriptLoad()
{
	print( "Free view camera script loaded" );
	print( "------------------------------" );
	print( "Type '/cam' to enable/disable the free view camera." );
	print( "Use W, A, S and D keys to move the camera and arrow keys to rotate it." );
	print( "Hold 'Shift' to increase the camera speed." );
	
	pCamera = CCamera();
	
	pLabel = GUILabel( VectorScreen( ScreenWidth - 225, ScreenHeight - 25 ), ScreenSize( 225, 25 ), "Camera pos: 0.0, 0.0, 0.0" );
	if ( pLabel )
	{
		pLabel.Visible = false;
		pLabel.TextColour = Colour( 255, 255, 255 );
		pLabel.FontSize = 9;
		pLabel.FontName = "Arial";
		
		AddGUILayer( pLabel );
		SendGUILayerToFront( pLabel );
	}
	
	return 1;
}

function onClientCommand( cmd, text )
{
	if ( cmd == "cam" )
	{
		if ( !pCamera.IsEnabled() )
		{
			pCamera.Enable();
			Message( "Free view camera enabled. Type '/cam' to disable." );
		}
		else
		{
			pCamera.Disable();
			Message( "Free view camera disabled." );
		}
	}
	
	return 1;
}

function onClientRender()
{
	if ( pCamera.IsEnabled() ) pCamera.Process();
	
	return 1;
}

// The following events protect the player from getting hurt while they are in free view mode.
function onClientShot( attacker, weapon, bodypart )
{
	if ( pCamera.IsEnabled() ) return 0;
	return 1;
}

function onClientHurt( entity )
{
	if ( pCamera.IsEnabled() ) return 0;
	return 1;
}
